#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "libpoc.h"

#define COL_LIGHT_BLUE  "\033[1;34m"
#define COL_LIGHT_RED   "\033[1;31m"
#define COL_RESET       "\033[0m"
#define COL_WHITE       "\033[1;37m"

#define ICON(c, x)      COL_WHITE "[" c x COL_WHITE "]" COL_RESET " "
#define ICON_INFO       ICON(COL_LIGHT_BLUE, "+")
#define ICON_ERROR      ICON(COL_LIGHT_RED, "!")

void
log_info(const char *fmt, ...)
{
    va_list args;

    fputs(ICON_INFO, stdout);

    va_start(args, fmt);
    vfprintf(stdout, fmt, args);
    va_end(args);

    fputs("\n", stdout);

    fflush(stdout);
}

void
log_hexdump(const void *data, size_t datasz, const char *fmt, ...)
{
    va_list args;
    const unsigned char *b;
    size_t n;
    size_t nascii;
    char ascii[8 + 1 + 8 + 1];
    size_t remainder;

    fputs(ICON_INFO, stdout);

    if (fmt) {
        va_start(args, fmt);
        vfprintf(stdout, fmt, args);
        va_end(args);

        fprintf(stdout, " (%zu bytes):\n", datasz);
    } else {
        fprintf(stdout, "%zu bytes:\n", datasz);
    }

    b = data;
    bzero(ascii, sizeof(ascii));
    nascii = 0;

    for (n = 0; n < datasz; ++n) {
        if (!(n & 0xf)) {
            fprintf(stdout, "%08zx: ", n);
        }

        fprintf(stdout, "%02x ", b[n]);
        ascii[nascii++] = isprint(b[n]) ? b[n] : '.';

        switch ((n + 1) & 0xf) {
        case 8:
            if (datasz > n + 1) {
                ascii[nascii++] = ' ';
                fputs(" ", stdout);
            }
            break;

        case 0:
            fprintf(stdout, " |%s|\n", ascii);

            bzero(ascii, sizeof(ascii));
            nascii = 0;
            break;
        }
    }

    remainder = 16 - (datasz & 0xf);
    if (remainder < 16) {
        while (remainder--) {
            fputs((remainder == 7) ? "    " : "   ", stdout);
        }

        fprintf(stdout, " |%s|\n", ascii);
    }

    fflush(stdout);
}

void
log_progress(const char *fmt, ...)
{
    va_list args;

    fputs("\r", stdout);
    fputs(ICON_INFO, stdout);

    va_start(args, fmt);
    vfprintf(stdout, fmt, args);
    va_end(args);

    fflush(stdout);
}

void
log_progress_complete(void)
{
    fputs("\n", stdout);
    fflush(stdout);
}

void
log_error(const char *fmt, ...)
{
    va_list args;

    fputs(ICON_ERROR, stderr);

    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);

    fputs("\n", stderr);
}

void
log_fatal(const char *fmt, ...)
{
    va_list args;

    fputs(ICON_ERROR, stderr);

    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);

    fputs("\n", stderr);
    exit(1);
}

void
log_fatal_errno(const char *fmt, ...)
{
    va_list args;

    fputs(ICON_ERROR, stderr);

    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
    va_end(args);

    fprintf(stderr, ": %s\n", strerror(errno));
    exit(1);
}

void
banner(void)
{
    log_info("%s", poc_title);
    log_info("By %s", poc_author);
    log_info("---");
}
